@@ -1,3 +1,7 @@
+// Copyright 2014 The Emscripten Authors.  All rights reserved.
+// Emscripten is available under two separate licenses, the MIT license and the
+// University of Illinois/NCSA Open Source License.  Both these licenses can be
+// found in the LICENSE file.
 
 var MAGIC = 0;
 Math.random = function() {
@@ -15,7 +19,7 @@ Module['thisProgram'] = 'thisProgram'; // for consistency between different buil
 
 function hashMemory(id) {
   var ret = 0;
-  var len = Math.max(DYNAMICTOP, STATICTOP);
+  var len = Math.max(HEAP32[DYNAMICTOP_PTR>>2], STATICTOP);
   for (var i = 0; i < len; i++) {
     ret = (ret*17 + HEAPU8[i])|0;
   }
