@@ -70,6 +70,7 @@ namespace emscripten {
             bool _emval_strictly_equals(EM_VAL first, EM_VAL second);
             bool _emval_greater_than(EM_VAL first, EM_VAL second);
             bool _emval_less_than(EM_VAL first, EM_VAL second);
+            bool _emval_not(EM_VAL object);
 
             EM_VAL _emval_call(
                 EM_VAL value,
@@ -97,6 +98,7 @@ namespace emscripten {
             bool _emval_instanceof(EM_VAL object, EM_VAL constructor);
             bool _emval_in(EM_VAL item, EM_VAL object);
             bool _emval_delete(EM_VAL object, EM_VAL property);
+            bool _emval_throw(EM_VAL object);
         }
 
         template<const char* address>
@@ -282,7 +284,7 @@ namespace emscripten {
     class val {
     public:
         // missing operators:
-        // * ! ~ - + ++ --
+        // * ~ - + ++ --
         // * * / %
         // * + -
         // * << >> >>>
@@ -426,18 +428,22 @@ namespace emscripten {
             return internal::_emval_greater_than(handle, v.handle);
         }
 
-        bool operator>= (const val& v) const {
+        bool operator>=(const val& v) const {
             return (*this > v) || (*this == v);
         }
 
-        bool operator< (const val& v) const {
+        bool operator<(const val& v) const {
             return internal::_emval_less_than(handle, v.handle);
         }
 
-        bool operator<= (const val& v) const {
+        bool operator<=(const val& v) const {
             return (*this < v) || (*this == v);
         }
 
+        bool operator!() const {
+            return internal::_emval_not(handle);
+        }
+
         template<typename... Args>
         val new_(Args&&... args) const {
             return internalCall(internal::_emval_new,std::forward<Args>(args)...);
@@ -511,6 +517,10 @@ namespace emscripten {
             return internal::_emval_delete(handle, val(property).handle);
         }
 
+        void throw_() const {
+            internal::_emval_throw(handle);
+        }
+
     private:
         // takes ownership, assumes handle already incref'd
         explicit val(internal::EM_VAL handle)
