@@ -12,6 +12,9 @@
 #include <emscripten/emscripten.h>
 #include <math.h>
 
+// Uncomment the following and clear the cache with emcc --clear-cache to rebuild this file to enable internal debugging.
+// #define FETCH_DEBUG
+
 struct __emscripten_fetch_queue
 {
 	emscripten_fetch_t **queuedOperations;
@@ -44,8 +47,10 @@ void emscripten_proxy_fetch(emscripten_fetch_t *fetch)
 	__emscripten_fetch_queue *queue = _emscripten_get_fetch_queue();
 //	TODO handle case when queue->numQueuedItems >= queue->queueSize
 	queue->queuedOperations[queue->numQueuedItems++] = fetch;
+#ifdef FETCH_DEBUG
 	EM_ASM(console.log('Queued fetch to fetch-worker to process. There are now ' + $0 + ' operations in the queue.'),
 		queue->numQueuedItems);
+#endif
 	// TODO: mutex unlock
 }
 
@@ -67,8 +72,10 @@ emscripten_fetch_t *emscripten_fetch(emscripten_fetch_attr_t *fetch_attr, const
 	const bool isMainBrowserThread = emscripten_is_main_browser_thread() != 0;
 	if (isMainBrowserThread && synchronous && (performXhr || readFromIndexedDB || writeToIndexedDB))
 	{
+#ifdef FETCH_DEBUG
 		EM_ASM(err('emscripten_fetch("' + Pointer_stringify($0) + '") failed! Synchronous blocking XHRs and IndexedDB operations are not supported on the main browser thread. Try dropping the EMSCRIPTEN_FETCH_SYNCHRONOUS flag, or run with the linker flag --proxy-to-worker to decouple main C runtime thread from the main browser thread.'), 
 			url);
+#endif
 		return 0;
 	}
 
@@ -162,23 +169,34 @@ EMSCRIPTEN_RESULT emscripten_fetch_wait(emscripten_fetch_t *fetch, double timeou
 	uint32_t proxyState = emscripten_atomic_load_u32(&fetch->__proxyState);
 	if (proxyState == 2) return EMSCRIPTEN_RESULT_SUCCESS; // already finished.
 	if (proxyState != 1) return EMSCRIPTEN_RESULT_INVALID_PARAM; // the fetch should be ongoing?
-// #ifdef FETCH_DEBUG
+#ifdef FETCH_DEBUG
 	EM_ASM(console.log('fetch: emscripten_fetch_wait..'));
-// #endif
+#endif
 	// TODO: timeoutMsecs is currently ignored. Return EMSCRIPTEN_RESULT_TIMED_OUT on timeout.
 	while(proxyState == 1/*sent to proxy worker*/)
 	{
-		emscripten_futex_wait(&fetch->__proxyState, proxyState, 100 /*TODO HACK:Sleep sometimes doesn't wake up?*/);//timeoutMsecs);
-		proxyState = emscripten_atomic_load_u32(&fetch->__proxyState);
+		if (!emscripten_is_main_browser_thread())
+		{
+			emscripten_futex_wait(&fetch->__proxyState, proxyState, 100 /*TODO HACK:Sleep sometimes doesn't wake up?*/);//timeoutMsecs);
+			proxyState = emscripten_atomic_load_u32(&fetch->__proxyState);
+		}
+		else 
+		{
+			EM_ASM({ console.error('fetch: emscripten_fetch_wait failed: main thread cannot block to wait for long periods of time! Migrate the application to run in a worker to perform synchronous file IO, or switch to using asynchronous IO.') });
+			return EMSCRIPTEN_RESULT_FAILED;
+		}
 	}
-// #ifdef FETCH_DEBUG
+#ifdef FETCH_DEBUG
 	EM_ASM(console.log('fetch: emscripten_fetch_wait done..'));
-// #endif
+#endif
 
 	if (proxyState == 2) return EMSCRIPTEN_RESULT_SUCCESS;
 	else return EMSCRIPTEN_RESULT_FAILED;
 #else
+
+#ifdef FETCH_DEBUG
 	EM_ASM(console.error('fetch: emscripten_fetch_wait is not available when building without pthreads!'));
+#endif
 	return EMSCRIPTEN_RESULT_FAILED;
 #endif
 }
