@@ -1,21 +1,28 @@
+#!/usr/bin/env python2
 # Copyright 2016 The Emscripten Authors.  All rights reserved.
 # Emscripten is available under two separate licenses, the MIT license and the
 # University of Illinois/NCSA Open Source License.  Both these licenses can be
 # found in the LICENSE file.
 
-'''
-Tries to evaluate global constructors, applying their effects ahead of time.
+"""Tries to evaluate global constructors, applying their effects ahead of time.
 
-This is an LTO-like operation, and to avoid parsing the entire tree (we might fail to parse a massive project, we operate on the text in python.
-'''
+This is an LTO-like operation, and to avoid parsing the entire tree (we might
+fail to parse a massive project, we operate on the text in python.
+"""
 
-import os, sys, json, subprocess, time
+import json
+import logging
+import os
+import subprocess
+import sys
+import time
 
 sys.path.insert(1, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
 
-from tools import shared, js_optimizer
+from tools import shared, js_optimizer, jsrun
 from tools.tempfiles import try_delete
 
+
 js_file = sys.argv[1]
 binary_file = sys.argv[2] # mem init for js, wasm binary for wasm
 total_memory = int(sys.argv[3])
@@ -32,9 +39,11 @@ config = shared.Configuration()
 
 # helpers
 
+
 def get_asm(js):
   return js[js.find(js_optimizer.start_asm_marker):js.find(js_optimizer.end_asm_marker)]
 
+
 def find_ctors(js):
   ctors_start = js.find('__ATINIT__.push(')
   if ctors_start < 0:
@@ -44,20 +53,24 @@ def find_ctors(js):
   ctors_end += 3
   return (ctors_start, ctors_end)
 
+
 def find_ctors_data(js, num):
   ctors_start, ctors_end = find_ctors(js)
   assert ctors_start > 0
   ctors_text = js[ctors_start:ctors_end]
   all_ctors = [ctor for ctor in ctors_text.split(' ') if ctor.endswith('()') and not ctor == 'function()' and '.' not in ctor]
   all_ctors = [ctor.replace('()', '') for ctor in all_ctors]
+  if shared.Settings.WASM_BACKEND:
+    assert all(ctor.startswith('_') for ctor in all_ctors)
+    all_ctors = [ctor[1:] for ctor in all_ctors]
   assert len(all_ctors)
   ctors = all_ctors[:num]
   return ctors_start, ctors_end, all_ctors, ctors
 
+
 def eval_ctors_js(js, mem_init, num):
 
   def kill_func(asm, name):
-    before = len(asm)
     asm = asm.replace('function ' + name + '(', 'function KILLED_' + name + '(', 1)
     return asm
 
@@ -65,13 +78,13 @@ def eval_ctors_js(js, mem_init, num):
     before = len(asm)
     asm = asm.replace('function ', ' ' + func + '\nfunction ', 1)
     assert len(asm) > before
-    name = func[func.find(' ')+1 : func.find('(')]
+    name = func[func.find(' ') + 1:func.find('(')]
     asm = asm.replace('return {', 'return { ' + name + ': ' + name + ',')
     return asm
 
   # Find the global ctors
   ctors_start, ctors_end, all_ctors, ctors = find_ctors_data(js, num)
-  shared.logging.debug('trying to eval ctors: ' + ', '.join(ctors))
+  logging.debug('trying to eval ctors: ' + ', '.join(ctors))
   # Find the asm module, and receive the mem init.
   asm = get_asm(js)
   assert len(asm)
@@ -254,19 +267,22 @@ console.log(JSON.stringify([numSuccessful, Array.prototype.slice.call(heap.subar
     err_file_handle = open(err_file, 'w')
     proc = subprocess.Popen(shared.NODE_JS + [temp_file], stdout=out_file_handle, stderr=err_file_handle, universal_newlines=True)
     try:
-      shared.jsrun.timeout_run(proc, timeout=10, full_output=True, throw_on_failure=False)
+      jsrun.timeout_run(proc, timeout=10, full_output=True, throw_on_failure=False)
     except Exception as e:
-      if 'Timed out' not in str(e): raise e
-      shared.logging.debug('ctors timed out\n')
+      if 'Timed out' not in str(e):
+        raise
+      logging.debug('ctors timed out\n')
       return (0, 0, 0, 0)
-    finally:
+    if shared.WINDOWS:
       time.sleep(0.5) # On Windows, there is some kind of race condition with Popen output stream related functions, where file handles are still in use a short period after the process has finished.
-      out_file_handle.close()
-      err_file_handle.close()
-      out_result = read_and_delete(out_file)
-      err_result = read_and_delete(err_file)
+    out_file_handle.close()
+    err_file_handle.close()
+    out_result = read_and_delete(out_file)
+    err_result = read_and_delete(err_file)
     if proc.returncode != 0:
-      shared.logging.debug('unexpected error while trying to eval ctors:\n' + out_result + '\n' + err_result)
+      # TODO(sbc): This should never happen under normal circumstances.
+      # switch to exit_with_error once we fix https://github.com/kripken/emscripten/issues/7463
+      logging.debug('unexpected error while trying to eval ctors:\n' + out_result + '\n' + err_result)
       return (0, 0, 0, 0)
 
   # out contains the new mem init and other info
@@ -274,7 +290,7 @@ console.log(JSON.stringify([numSuccessful, Array.prototype.slice.call(heap.subar
   mem_init = bytes(bytearray(mem_init_raw))
   total_ctors = len(all_ctors)
   if num_successful < total_ctors:
-    shared.logging.debug('not all ctors could be evalled, something was used that was not safe (and therefore was not defined, and caused an error):\n========\n' + err_result + '========')
+    logging.debug('not all ctors could be evalled, something was used that was not safe (and therefore was not defined, and caused an error):\n========\n' + err_result + '========')
   # Remove the evalled ctors, add a new one for atexits if needed, and write that out
   if len(ctors) == total_ctors and len(atexits) == 0:
     new_ctors = ''
@@ -288,21 +304,25 @@ console.log(JSON.stringify([numSuccessful, Array.prototype.slice.call(heap.subar
   js = js[:ctors_start] + new_ctors + js[ctors_end:]
   return (num_successful, js, mem_init, ctors)
 
+
 def eval_ctors_wasm(js, wasm_file, num):
   ctors_start, ctors_end, all_ctors, ctors = find_ctors_data(js, num)
   cmd = [os.path.join(binaryen_bin, 'wasm-ctor-eval'), wasm_file, '-o', wasm_file, '--ctors=' + ','.join(ctors)]
   if debug_info:
     cmd += ['-g']
-  shared.logging.debug('wasm ctor cmd: ' + str(cmd))
+  logging.debug('wasm ctor cmd: ' + str(cmd))
   proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)
   try:
-    err = shared.jsrun.timeout_run(proc, timeout=10, full_output=True, throw_on_failure=False)
+    err = jsrun.timeout_run(proc, timeout=10, full_output=True, throw_on_failure=False)
   except Exception as e:
-    if 'Timed out' not in str(e): raise e
-    shared.logging.debug('ctors timed out\n')
+    if 'Timed out' not in str(e):
+      raise
+    logging.debug('ctors timed out\n')
     return 0, js
+  if proc.returncode != 0:
+    shared.exit_with_error('unexpected error while trying to eval ctors:\n' + err)
   num_successful = err.count('success on')
-  shared.logging.debug(err)
+  logging.debug(err)
   if len(ctors) == num_successful:
     new_ctors = ''
   else:
@@ -313,21 +333,22 @@ def eval_ctors_wasm(js, wasm_file, num):
   js = js[:ctors_start] + new_ctors + js[ctors_end:]
   return num_successful, js
 
+
 # main
-if __name__ == '__main__':
+def main():
   js = open(js_file).read()
   ctors_start, ctors_end = find_ctors(js)
   if ctors_start < 0:
-    shared.logging.debug('ctor_evaller: no ctors')
+    logging.debug('ctor_evaller: no ctors')
     sys.exit(0)
 
-  ctors_text = js[ctors_start:ctors_end];
+  ctors_text = js[ctors_start:ctors_end]
   if ctors_text.count('(') == 1:
-    shared.logging.debug('ctor_evaller: push, but no ctors')
+    logging.debug('ctor_evaller: push, but no ctors')
     sys.exit(0)
 
   num_ctors = ctors_text.count('function()')
-  shared.logging.debug('ctor_evaller: %d ctors, from |%s|' % (num_ctors, ctors_text))
+  logging.debug('ctor_evaller: %d ctors, from |%s|' % (num_ctors, ctors_text))
 
   if not wasm:
     # js path
@@ -339,18 +360,18 @@ if __name__ == '__main__':
 
     # find how many ctors we can remove, by bisection (if there are hundreds, running them sequentially is silly slow)
 
-    shared.logging.debug('ctor_evaller: trying to eval %d global constructors' % num_ctors)
+    logging.debug('ctor_evaller: trying to eval %d global constructors' % num_ctors)
     num_successful, new_js, new_mem_init, removed = eval_ctors_js(js, mem_init, num_ctors)
     if num_successful == 0:
-      shared.logging.debug('ctor_evaller: not successful')
+      logging.debug('ctor_evaller: not successful')
       sys.exit(0)
 
-    shared.logging.debug('ctor_evaller: we managed to remove %d ctors' % num_successful)
+    logging.debug('ctor_evaller: we managed to remove %d ctors' % num_successful)
     if num_successful == num_ctors:
       js = new_js
       mem_init = new_mem_init
     else:
-      shared.logging.debug('ctor_evaller: final execution')
+      logging.debug('ctor_evaller: final execution')
       check, js, mem_init, removed = eval_ctors_js(js, mem_init, num_successful)
       assert check == num_successful
     open(js_file, 'w').write(js)
@@ -358,12 +379,12 @@ if __name__ == '__main__':
 
     # Dead function elimination can help us
 
-    shared.logging.debug('ctor_evaller: eliminate no longer needed functions after ctor elimination')
+    logging.debug('ctor_evaller: eliminate no longer needed functions after ctor elimination')
     # find exports
     asm = get_asm(open(js_file).read())
     exports_start = asm.find('return {')
     exports_end = asm.find('};', exports_start)
-    exports_text = asm[asm.find('{', exports_start) + 1 : exports_end]
+    exports_text = asm[asm.find('{', exports_start) + 1:exports_end]
     exports = [x.split(':')[1].strip() for x in exports_text.replace(' ', '').split(',')]
     for r in removed:
       assert r in exports, 'global ctors were exported'
@@ -377,15 +398,18 @@ if __name__ == '__main__':
     reachable = shared.Building.calculate_reachable_functions(js_file, exports, can_reach=False)['reachable']
     for r in removed:
       assert r not in reachable, 'removed ctors must NOT be reachable'
-    shared.Building.js_optimizer(js_file, ['removeFuncs'], extra_info={ 'keep': reachable }, output_filename=js_file)
+    shared.Building.js_optimizer(js_file, ['removeFuncs'], extra_info={'keep': reachable}, output_filename=js_file)
   else:
     # wasm path
     wasm_file = binary_file
-    shared.logging.debug('ctor_evaller (wasm): trying to eval %d global constructors' % num_ctors)
+    logging.debug('ctor_evaller (wasm): trying to eval %d global constructors' % num_ctors)
     num_successful, new_js = eval_ctors_wasm(js, wasm_file, num_ctors)
     if num_successful == 0:
-      shared.logging.debug('ctor_evaller: not successful')
+      logging.debug('ctor_evaller: not successful')
       sys.exit(0)
-    shared.logging.debug('ctor_evaller: we managed to remove %d ctors' % num_successful)
+    logging.debug('ctor_evaller: we managed to remove %d ctors' % num_successful)
     open(js_file, 'w').write(new_js)
 
+
+if __name__ == '__main__':
+  sys.exit(main())
